#ifndef TYPON_CORE_ROOT_HPP_INCLUDED
#define TYPON_CORE_ROOT_HPP_INCLUDED

#include <atomic>
#include <coroutine>
#include <cstdint>

//
#include <typon/core/continuation.hpp>
#include <typon/core/result.hpp>
#include <typon/core/scheduler.hpp>


namespace typon
{

  struct [[nodiscard]] Root
  {
    struct promise_type;

    std::coroutine_handle<promise_type> _coroutine;

    ~Root()
    {
      _coroutine.destroy();
    }

    struct promise_type : Result<void>
    {
      Continuation::Data _data;

      promise_type() noexcept
        : _data(std::coroutine_handle<promise_type>::from_promise(*this))
      {}

      Root get_return_object() noexcept
      {
        return { std::coroutine_handle<promise_type>::from_promise(*this) };
      }

      std::suspend_always initial_suspend() noexcept
      {
        return {};
      }

      auto final_suspend() noexcept
      {
        struct awaitable : std::suspend_always
        {
          void await_suspend(std::coroutine_handle<promise_type> coroutine) noexcept
          {
            // !! bad racy
            coroutine.promise()._data._n.store(0, std::memory_order_release);
            coroutine.promise()._data._n.notify_one();
          }
        };

        return awaitable {};
      }
    };

    void call() &&
    {
      Scheduler::get();
      Scheduler::schedule(_coroutine);
      _coroutine.promise()._data._n.wait(Continuation::Data::UMAX, std::memory_order_acquire);
    }
  };

}

#endif // TYPON_ROOT_HPP_INCLUDED
