#ifndef TYPON_FUNDAMENTAL_OPTIONAL_HPP_INCLUDED
#define TYPON_FUNDAMENTAL_OPTIONAL_HPP_INCLUDED

#include <type_traits>


namespace typon::fdt
{

  template <typename T>
    requires std::is_trivially_copyable_v<T>
  struct optional
  {
    template <unsigned char I>
    using state = std::integral_constant<unsigned char, I>;

    unsigned char _state;
    union
    {
      T _value;
    };

    optional() noexcept : _state(0) {}

    template <unsigned char I>
    optional(state<I> state) noexcept : _state(state() << 1 & (~1)) {}

    optional(T value) noexcept : _state(1), _value(value) {}

    template <unsigned char I>
    optional(T value, state<I> state) noexcept
      : _state((state() << 1) | 1)
      , _value(value)
    {}

    ~optional()
    {
      if (_state & 1)
      {
        std::destroy_at(std::addressof(_value));
      }
    }

    operator bool() noexcept
    {
      return _state & 1;
    }

    template <unsigned char I>
    bool match(state<I> state) noexcept
    {
      return state() == _state >> 1;
    }

    template <unsigned char I>
    void set_state(state<I> state) noexcept
    {
      _state = (state() << 1) | (_state & 1);
    }

    T * operator->() noexcept
    {
      return std::addressof(_value);
    }

    T & operator*() noexcept
    {
      return _value;
    }
  };

}


#endif // TYPON_FUNDAMENTAL_OPTIONAL_HPP_INCLUDED
